;; i!
;;
;; ( x addr -- )
;;
;; store cell to instruction memory

link_
.db "i!",0,0

istore:
  rcall over
  rcall over
  rcall spmbuf
  dup_
  rcall pageload
  dup_
  rcall spmerase
  dup_
  rcall spmwrite
  rcall spmrww
  drop_
  drop_
  ret
  

;; pageload
;;
;; ( addr -- )

pageload:
  ; save complete address for later
  dup_
  rcall tor

  ; get beginning of page
  and_ pagemask

  ; loop counter (in words)
  lit_ pagesize
  rcall swapp

pageload_loop:
  rcall over
  adiw tos,0
  drop_
  breq pageload_done

  ; don't update temp buffer if it's the
  ; value we're changing. each location
  ; can only be written once.
  dup_
  rcall rfetch
  rcall notequal
  adiw tos,0
  drop_
  breq pageload_dontcopy

pageload_copy:
  dup_
  rcall ifetch
  rcall over
  rcall spmbuf

pageload_dontcopy:
  oneplus_
  rcall swapp
  oneminus_
  rcall swapp
  rjmp pageload_loop

pageload_done:
  drop_
  drop_
  rcall rfrom
  drop_
  ret


;; dospm
;;
;; ( spmcsr x addr -- )
;;
;; execute spm instruction

link_
.db "dospm",0

dospm:
  ; disable interrupts if enabled,
  ; store status register
  in_ temp0,SREG
  cli

  movw z,tos
  szl
  drop_

  movw r0,tos
  drop_

  out_ SPMCSR,tosl
  spm
  drop_

dospm_wait:
  ; wait until done
  sbic_ SPMCSR,SPMEN,temp1
  rjmp dospm_wait

  ; restore status register
  out_ SREG,temp0
  ret


;; spmbuf
;;
;; ( x addr -- )

link_
.db "spmbuf",0,0

spmbuf:
  rcall tor
  rcall tor
  lit_ (1<<SPMEN)
  rcall rfrom
  rcall rfrom
  rcall dospm
  ret

;; spmerase
;;
;; ( addr -- )

link_
.db "spmerase",0,0

spmerase:
  and_ pagemask

  rcall tor
  lit_ (1<<PGERS|1<<SPMEN)
  lit_ 0
  rcall rfrom
  rcall dospm
  ret

;; spmwrite
;;
;; ( addr -- )

link_
.db "spmwrite",0,0

spmwrite:
  and_ pagemask
  
  rcall tor
  lit_ (1<<PGWRT|1<<SPMEN)
  lit_ 0
  rcall rfrom
  rcall dospm
  ret

;; spmrww
;;
;; ( -- )

link_
.db "spmrww",0,0

spmrww:
  lit_ (1<<RWWSRE|1<<SPMEN)
  lit_ 0
  lit_ 0
  rcall dospm
  ret

