;; um/mod
;;
;; ( ud u1 -- u2 u3 )
;;
;; unsigned 32/16 -> 16r16 divide

link_
.db "um/mod",0,0

umslashmod:
  ; load dividend
  ld temp2,y+
  ld temp3,y+
  ld temp0,y+
  ld temp1,y+

  ; logical shift left
  lsl temp0
  rol temp1
  rol temp2
  rol temp3
  
  ; set loop counter
  ldi xl,$10
  tst xl
  
umslashmod_loop:
  ; try subtracting divisor
  sub temp2,tosl
  sbc temp3,tosh
  
  brmi umslashmod_negative

umslashmod_positive:
  ; result was positive
  ; shift 1 in on the right
  sec
  rjmp umslashmod_shift

umslashmod_negative:
  ; result was negative
  ; add back divisor and
  ; shift 0 in on the right
  add temp2,tosl
  adc temp3,tosh
  clc

umslashmod_shift:
  ; do the shift
  rol temp0
  rol temp1
  rol temp2
  rol temp3
  
  dec xl
  brne umslashmod_loop

  ; shift remainder right
  lsr temp3
  ror temp2

  ; put remainder on stack
  st -y,temp3
  st -y,temp2

  ; put quotient on stack
  movw tos,temp0
  ret

  
link_
.db "um/mod2",0

umslashmod2:
  ; load dividend
  ld temp2,y+
  ld temp3,y+
  ld temp0,y+
  ld temp1,y+

  ; set loop counter
  ldi xl,$11

  clc

umslashmod2_loop:
  rol temp0
  rol temp1
  dec xl
  breq umslashmod2_done

  rol temp2
  rol temp3
  sub temp2,tosl
  sbc temp3,tosh
  brcc umslashmod2_positive

umslashmod2_negative:
  ; result was negative
  ; add back divisor and
  ; shift 0 in on the right
  add temp2,tosl
  adc temp3,tosh
  clc
  rjmp umslashmod2_loop
  
umslashmod2_positive:
  ; result was positive
  ; shift 1 in on the right
  sec
  rjmp umslashmod2_loop

umslashmod2_done:
  ; put remainder on stack
  st -y,temp3
  st -y,temp2

  ; put quotient on stack
  movw tos,temp0
  ret

  
link_
.db "m/mod",0

mslashmod:
  ; load dividend
  ld temp2,y+
  ld temp3,y+
  ld temp0,y+
  ld temp1,y+

  ; xor signs
  mov xh,temp3
  eor xh,tosh

mslashmod_abs_dividend:
  ; get absolute value of dividend
  tst temp3
  brpl mslashmod_abs_divisor
  com temp0
  com temp1
  com temp2
  com temp3
  subi temp0,-1
  sbci temp1,0
  sbci temp2,0
  sbci temp3,0

mslashmod_abs_divisor:
  tst tosh
  brpl mslashmod_start
  com tosl
  com tosh
  adiw tos,1
  
mslashmod_start:  
  ; set loop counter
  ldi xl,$11

  clc

mslashmod_loop:
  rol temp0
  rol temp1
  dec xl
  breq mslashmod_sign

  rol temp2
  rol temp3
  sub temp2,tosl
  sbc temp3,tosh
  brcc mslashmod_positive

mslashmod_negative:
  ; result was negative
  ; add back divisor and
  ; shift 0 in on the right
  add temp2,tosl
  adc temp3,tosh
  clc
  rjmp mslashmod_loop
  
mslashmod_positive:
  ; result was positive
  ; shift 1 in on the right
  sec
  rjmp mslashmod_loop

mslashmod_sign:
  ; change sign of quotient if signs of
  ; dividend and divsor were different
  tst xh
  brpl mslashmod_done
  com temp0
  com temp1
  subi temp0,-1
  sbci temp1,0

mslashmod_done:  
  ; put remainder on stack
  st -y,temp3
  st -y,temp2

  ; put quotient on stack
  movw tos,temp0
  ret
  
